from typing import Dict, Optional

from attrs import define


@define
class HttpClientConfiguration:
    headers: Optional[Dict[str, str]] = None
    cookies: Optional[Dict[str, str]] = None
    verify: bool = True
    trust_env: bool = True



import attrs
import pytest
from httpx import URL, Client
from pytest import FixtureRequest

from .http_client_configuration import HttpClientConfiguration
from .page_checker import PageChecker


@pytest.fixture
def http_client_config():
    return HttpClientConfiguration()


@pytest.fixture
def http_client(http_client_config: HttpClientConfiguration):
    with Client(**attrs.asdict(http_client_config)) as client:
        yield client


@pytest.fixture
def check(request: FixtureRequest, http_client: Client):
    if not hasattr(request.module, "host"):
        raise ValueError("'host' should be set in test file")
    host = getattr(request.module, "host")
    scheme = "https" if not hasattr(request.module, "scheme") else getattr(request.module, "scheme")
    port = None if not hasattr(request.module, "port") else int(getattr(request.module, "port"))
    return PageChecker(http_client=http_client, base_url=URL(scheme=scheme, host=host, port=port))



from httpx import Client, Response

from .assert_element_checker_base import AssertElementCheckerBase
from .assertion_attribute_base import AssertionAttributeBase


class _StatusCodeChecker(AssertElementCheckerBase[int]):
    def check(self, http_client: Client, response: Response, negative: bool = False) -> None:
        if self.value:
            if (self.value != response.status_code) ^ negative:
                raise AssertionError(
                    self._make_message(
                        info=f"'{self.value}'", check_type="status_code", url=str(response.url), negative=negative
                    )
                )


class StatusCodeAssertion(AssertionAttributeBase):
    _checker_type = _StatusCodeChecker



import json
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional

from httpx import URL, Client, Response

from .assertions import Assertions, NegativeAssertions


class PageChecker:
    _base_url: URL
    _http_client: Client

    def __init__(self, http_client: Client, base_url: URL):
        self._base_url = base_url
        self._http_client = http_client

    def __call__(
        self,
        *,
        title: str,
        path: str,
        base_url: Optional[URL] = None,
        method: str = "GET",
        data: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None,
        follow_redirects: bool = False,
        should_find: Optional[Assertions] = None,
        should_not_find: Optional[NegativeAssertions] = None,
        **_: Any,
    ) -> None:
        """
        Call the target url with given arguments, then verify the response against given rules
        Return None if success, otherwise raise AssertionError
        """
        base_url = base_url or self._base_url
        response = self._http_client.request(
            method=method,
            url=base_url.copy_with(raw_path=path.encode("ascii")),
            data=data,
            headers=headers,
            cookies=cookies,
            follow_redirects=follow_redirects,
        )
        try:
            if should_find:
                should_find.check_assertions(http_client=self._http_client, response=response)
            if should_not_find:
                should_not_find.check_assertions(http_client=self._http_client, response=response)
        except AssertionError as exc:
            file_name = self._dump_response(response=response)
            msg = f"{title} - {str(exc)} - please check file '{file_name}'"
            raise AssertionError(msg) from exc

    @staticmethod
    def _dump_response(response: Response) -> str:
        """
        Save the response's essential contents into a temporay file
        Return the file name
        """
        result: Dict[str, Any] = {
            "status_code": response.status_code,
            "content": response.text,
            "headers": dict(response.headers),
            "cookies": dict(response.cookies),
        }
        with NamedTemporaryFile(mode="wt", delete=False) as f:
            json.dump(result, f)
            return f.name



from typing import ClassVar, Type
from weakref import WeakKeyDictionary

from httpx import Client, Response

from .assert_element_checker_base import AssertElementCheckerBase


class AssertionAttributeBase:
    _checker_type: Type[AssertElementCheckerBase]
    # if the owner instance of the descriptor is deleted, we should also remove the its entry from assertion_instances
    # that's why we use WeakKeyDictionary
    # it groups the checkers by owner : {owner_instance: [checker_instance, ...]}
    checkers: ClassVar[WeakKeyDictionary] = WeakKeyDictionary()

    def __get__(self, instance, owner):
        if not hasattr(instance, self._private_name):
            return self._checker_type()
        return getattr(instance, self._private_name)

    def __set_name__(self, owner, name):
        self._private_name = f"_{name}"

    def __set__(self, instance, value):
        checker = self._checker_type(value)
        if instance not in self.checkers:
            self.checkers.setdefault(instance, [])
        self.checkers[instance].append(checker)
        setattr(instance, self._private_name, checker)


def check_all(instance, http_client: Client, response: Response, negative: bool):
    checker: AssertElementCheckerBase
    for checker in AssertionAttributeBase.checkers[instance]:
        checker.check(http_client, response, negative)



from typing import Generic, Optional, TypeVar

from httpx import Client, Response

T = TypeVar("T")


class AssertElementCheckerBase(Generic[T]):
    value: Optional[T]

    def __init__(self, value: Optional[T] = None):
        self.value = value

    def check(self, http_client: Client, response: Response, negative: bool = False):
        raise NotImplementedError

    @staticmethod
    def _make_message(info: str, check_type: str, url: str, negative: bool) -> str:
        return f"{info}{'' if negative else ' not'} " f"found in {check_type} on page '{url}'"



from typing import Mapping, Optional, Sequence

from httpx import Client, Response

from .assertion_elements.assertion_attribute_base import check_all
from .assertion_elements.content_assertion import ContentAssertion
from .assertion_elements.cookies_assertion import Cookie, CookiesAssertion
from .assertion_elements.headers_assertion import HeadersAssertion
from .assertion_elements.status_code_assertion import StatusCodeAssertion


class _AssertionsBase:
    _negative: bool = False
    status_code = StatusCodeAssertion()
    content = ContentAssertion()
    headers = HeadersAssertion()
    cookies = CookiesAssertion()

    def __init__(
        self,
        status_code: Optional[int] = None,
        content: Optional[Sequence[str]] = None,
        headers: Optional[Mapping[str, str]] = None,
        cookies: Optional[Sequence[Cookie]] = None,
    ):
        self.status_code = status_code
        self.content = content
        self.headers = headers
        self.cookies = cookies

    def check_assertions(self, http_client: Client, response: Response) -> None:
        check_all(instance=self, http_client=http_client, response=response, negative=self._negative)


class Assertions(_AssertionsBase):
    _negative: bool = False


class NegativeAssertions(_AssertionsBase):
    _negative: bool = True



import re
from typing import Sequence

from httpx import Client, Response

from .assert_element_checker_base import AssertElementCheckerBase
from .assertion_attribute_base import AssertionAttributeBase


class _ContentChecker(AssertElementCheckerBase[Sequence[str]]):
    def check(self, http_client: Client, response: Response, negative: bool = False) -> None:
        if self.value is None:
            return
        for content in self.value:
            if (re.search(content, response.text, re.MULTILINE) is None) ^ negative:
                raise AssertionError(
                    self._make_message(
                        info=f"'{self.value}'", check_type="content", url=str(response.url), negative=negative
                    )
                )


class ContentAssertion(AssertionAttributeBase):
    _checker_type = _ContentChecker



import re
from typing import Mapping

from httpx import Client, Response

from .assert_element_checker_base import AssertElementCheckerBase
from .assertion_attribute_base import AssertionAttributeBase


class _HeadersChecker(AssertElementCheckerBase[Mapping[str, str]]):
    def check(self, http_client: Client, response: Response, negative: bool = False) -> None:
        if self.value is None:
            return
        for header_key, header_value in self.value.items():
            # the header names in response are case-insensitive
            is_header_not_found = (
                header_key not in response.headers or re.search(header_value, response.headers[header_key]) is None
            )
            if is_header_not_found ^ negative:
                raise AssertionError(
                    self._make_message(
                        info=f"'{header_key}':'{header_value}'",
                        check_type="headers",
                        url=str(response.url),
                        negative=negative,
                    )
                )


class HeadersAssertion(AssertionAttributeBase):
    _checker_type = _HeadersChecker



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['http_testing', 'http_testing.assertion_elements']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=22.1.0,<23.0.0', 'httpx>=0.23.0,<0.24.0', 'pytest>=7.2.0,<8.0.0']

entry_points = \
{'pytest11': ['http_page_checker = http_testing.plugin']}

setup_kwargs = {
    'name': 'pytest-httptesting',
    'version': '0.1.2',
    'description': 'http_testing framework on top of pytest',
    'long_description': '# HTTP_TESTING\n\n## Description\nThis project aims to help to create e2e tests, by chaining http calls and verifications on target pages.\nIMPORTANT: Since only http calls involved, no javascript will run.\n\n## Concept\nThis project is built on pytest.\n\nEach .py file in test represents several tests scenario on one target site, we can provide the site\'s hostname\nas a local variable, the framework know how to construct http call from that.\n\nEach test function in the .py test file represent a scenario of test which is consisted by several steps. For example,\ntest user\'s account page, first, we authenticate the client by post user\'s name and password,\nthen access the account page to verify some values. Each step is described by calling the variable `check`\nwhich is a pytest fixture.\n\n## Tutorial\n### Install\n```bash\npip install pytest-httptesting\n```\n\n### Create test suite\n```python\n# test/test_example.py\n\nfrom http_testing.assertion_elements.cookies_assertion import Cookie\nfrom http_testing.assertions import Assertions, NegativeAssertions\nfrom http_testing.page_checker import PageChecker\n\nhost = "www.google.com"  # madatory: used in the `check` fixture\nscheme = "https"  # "https" by default\nport = None  # None by default\n\n\ndef test_scenario_one(check: PageChecker):\n    check(\n        title="Senario One",\n        path="/",\n        should_find=Assertions(\n            status_code=200,\n            content=["<title>Google</title>"],\n            headers={"Content-Type": "text/html; charset=ISO-8859-1"},\n            cookies=[Cookie(name="AEC", value_pattern=r".*")],\n        ),\n        should_not_find=NegativeAssertions(\n            status_code=400,\n            content=["groot"],\n            headers={"nooooo": ""},\n            cookies=[Cookie(name="nop", value_pattern="a")],\n        ),\n    )\n```\n\n### Run test\n```bash\n$ pytest --tb=no test  # traceback is disabled because it is not very useful to anayse the functional error\n============= test session starts =============\nplatform linux -- Python 3.8.10, pytest-7.2.0, pluggy-1.0.0\nrootdir: /home/xxx/workspace/http_testing\nplugins: anyio-3.6.2\ncollected 1 items\n\ntest/test_example.py .                                                                  [100%]\n\n=============  1 passed in 0.22s =============\n\n```\n\n### Debug\nIn case of error, a temporary file will be generated, as shown in the `test summary info`. It is a json file concluding\nresponse content, status code, headers and cookies.\n```bash\n$ pytest --tb=no test\n============= test session starts =============\nplatform linux -- Python 3.8.10, pytest-7.2.0, pluggy-1.0.0\nrootdir: /home/qile/workspace/http_testing\nplugins: anyio-3.6.2\ncollected 1 item\n\ntest/test_example.py F                                                                  [100%]\n\n============= short test summary info =============\nFAILED test/test_example.py::test_scenario_one - AssertionError: Senario One - \'Content-Typesssssss\':\'text/html; charset=ISO-8859-1\' not found in headers on page \'https://www.google.com/\' - please check file \'/tmp/tmpx7zb7h16\'\n============= 1 failed in 0.25s =============\n\n```\n\n### Advanced\n#### Customize the http client configuration\nWe can create a fixture `http_client_config` to override the default configuration, like adding headers and cookies to the http client.\n```python\n@pytest.fixture\ndef http_client_config():\n    return HttpClientConfiguration(\n        trust_env=False,\n        verify=False,\n        cookies={"cookie_1": "cookie_value_1"},\n        headers={"header_1": "header_value_1"},\n    )\n```\n',
    'author': 'HE Qile',
    'author_email': 'mr.qile@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/heqile/http_testing',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import re
from collections import defaultdict
from http.cookiejar import Cookie as HttpCookie
from typing import Mapping, Optional, Sequence

from attrs import define
from httpx import Client, Response

from .assert_element_checker_base import AssertElementCheckerBase
from .assertion_attribute_base import AssertionAttributeBase


@define
class Cookie:
    name: str
    value_pattern: str
    domain: Optional[str] = None
    path: Optional[str] = None
    secure: Optional[bool] = None
    expires: Optional[int] = None


class _CookiesChecker(AssertElementCheckerBase[Sequence[Cookie]]):
    def check(self, http_client: Client, response: Response, negative: bool = False) -> None:
        if self.value is None:
            return
        for cookie in self.value:
            # same cookie name can be on different sites, so use dict to group cookies list by name
            all_cookies = defaultdict(list)
            for cookiejar in http_client.cookies.jar:
                all_cookies[cookiejar.name].append(cookiejar)
            for cookiejar in response.cookies.jar:
                all_cookies[cookiejar.name].append(cookiejar)

            is_cookie_exist = self._is_cookie_match(target_cookie=cookie, all_cookies=all_cookies)
            if (not is_cookie_exist) ^ negative:
                raise AssertionError(
                    self._make_message(
                        info=f"'{cookie.name}':'{cookie.value_pattern}'",
                        check_type="cookies",
                        url=str(response.url),
                        negative=negative,
                    )
                )

    @staticmethod
    def _is_cookie_match(target_cookie: Cookie, all_cookies: Mapping[str, Sequence[HttpCookie]]):
        cookies = all_cookies.get(target_cookie.name)
        if cookies is None:
            return False
        for cookie in cookies:
            if cookie.value is None:
                continue
            if target_cookie.domain is not None and cookie.domain_specified and target_cookie.domain != cookie.domain:
                continue
            if target_cookie.path is not None and cookie.path_specified and target_cookie.path != cookie.path:
                continue
            if target_cookie.secure is not None and target_cookie.secure != cookie.secure:
                continue
            if target_cookie.expires is not None and target_cookie.expires != cookie.expires:
                continue
            if re.search(target_cookie.value_pattern, cookie.value) is not None:
                return True
        return False


class CookiesAssertion(AssertionAttributeBase):
    _checker_type = _CookiesChecker
